"""Tools to test codecs -- i.e. serialization/deserialization pairs

`validate_codec`: a function to test encoder/decoder pairs.

>>> from functools import partial
>>> import json
>>> validate_jsonability = partial(validate_codec, coder=json.dumps, decoder=json.loads)

You can jsonize lists and dicts

>>> assert validate_jsonability([1, 2, 3])
>>> assert validate_jsonability({'a': 1, 'b': {'c': [1, 2, 3]}})

You can't jsonize datetime objects

>>> from datetime import datetime
>>> validate_jsonability(datetime.now())
False

See `validate_codec` docs for more examples.

"""
import pickle


def return_object_itself(obj):
    return obj


def always_true(obj, decoded_obj):
    return True


def encode_and_decode(obj, coder=pickle.dumps, decoder=pickle.loads):
    encoded_obj = coder(obj)
    decoded_obj = decoder(encoded_obj)
    return decoded_obj


# TODO: Add doctests using caught_errors
# TODO: Discuss how the validate_codec can also solve decoding in a different env or
#  location
def validate_codec(
    obj,
    coder=pickle.dumps,
    decoder=pickle.loads,
    command=return_object_itself,
    comparison=always_true,
    caught_errors=(Exception,),
):
    """Validate a coder/decoder pair.

    Encodes the input `obj` with `coder`, then `decodes` the encoded object,
    returning `comparison(command(obj), command(decoded_obj))`.
    If an exception is raised, that is caught by `caught_errors`, `False` will be
    returned.

    :param obj: Object that should be serialized/encoded then deserialized/decoded
    :param coder: The serializer.
    :param decoder: The deserializer.
    :param command: Function to apply to both obj and decoded_obj before comparing
    :param comparison: Function that will be called on (obj, decoded_obj)
    :param caught_errors: The exception types to catch (and return False)
    :return: Whatever comparison returns

    Note: The `command` function is not necessary since a custom comparison function
    could be made to apply such a function to the obj and decoded_obj before comparing.
    We provide command though, because it allows the user to set a comparison function,
    but vary what is being compared.
    For instance, we may want to use `==` (i.e. `operator.eq`) as our comparison
    function, but apply this not to the objects themselves, but to some results
    of operating with or on them.
    For example, say you may not want to compare `func == decoded_func`, but instead
    `func(42) == decoded_func(42)`.
    In that case, you would specify `command=lambda f: f(42)`.

    To demo how `validate_codec` works, we'll import two classes, and print their
    code below. The reason we don't define them inline is because pickle can't
    be used inside a module where the type of the object to pickle is defined.

    >>> from tested.tests.objects_for_testing import A, B
    >>> import inspect
    >>> print(inspect.getsource(A))
    class A:
        def __init__(self, x):
            self.x = x
    <BLANKLINE>
        def __add__(self, other):
            return self.x + other
    <BLANKLINE>
    >>> print(inspect.getsource(B))
    class B(A):
        def __eq__(self, other):
            return self.x == other.x
    <BLANKLINE>

    Let's make a few objects to test with:

    >>> a_list = [1,2,3]
    >>> a_tuple = (1, 2, 3)
    >>> a = A(42)
    >>> b = B(42)

    Out-of-the-box, the only thing that `validate_codec` validates, is the ability of
    a coder/decoder pair to serialize, then deserialize an object.

    >>> assert validate_codec(a_list)
    >>> assert validate_codec(a)

    The default `coder` and `decoder` are `pickle.dumps` and `pickle.loads`
    respectively, so you wouldn't be able to pickle a lambda function, for instance:

    >>> assert not validate_codec(lambda x: x)

    The real usage of `validate_codec` though is to use `functools.partial` to fix the
    arguments to those that make sense for the codec and objects to be coded and decoded.

    Say for example, that we want to use equality `==` to compare the original object
    with the decoded one. In that case, we'd do this:

    >>> from functools import partial
    >>> from operator import eq
    >>>
    >>> equality_validator = partial(validate_codec, comparison=eq)
    >>> assert equality_validator(a_list)
    >>> assert not equality_validator(a)
    >>> assert equality_validator(b)

    The instance `a` of `A` didn't pass validation because by default, python objects
    compare on the basis of their `id`, which is a low level unique reference (
    integer) of an object.
    It worked with `a_list` and with `b` because these have their own definition of
    equality that bipasses `id`.

    If we wanted to force the comparison based on `id` anyway, we could do it like this:

    >>> id_equality_validator = partial(validate_codec, command=id, comparison=eq)
    >>> assert not id_equality_validator(a_list)
    >>> assert not id_equality_validator(a)
    >>> assert not id_equality_validator(b)

    The `id`-based comparison above was just meant for illustration.
    I can't think of any use case where it would be useful.
    On the other hand, the `command` argument has it's uses.

    Here is now the crux of the matter: Functional equivalence.

    No matter what anyone tells you, objects are useless in and of themselves.
    It's their use, their behavior, that matters.
    Therefore, when you decode an object, you're not really looking for the exact
    same object that was encoded.
    We've seen that in fact, from the point of view of low level concerns such as
    the `id`, you'll never get the exact same object.
    What you should expect to get though, is some equivalence from the behavior point
    of view.
    That is, you need to be able to rely on getting "the same" results when operating
    with the decoded object.

    This is where `command` and `comparison` come in.

    `command` specifies what operation(s) you want to be able to do on the decoded
    object, and `comparison` specifies how you compare the results of carrying out
    those same operations on the original object and the decoded one.

    Note that `A` (and therefore `B`) have been defined so that we can sum instances
    with a number:

    >>> a + 1
    43

    If what I really want out of my `A` and `B` instances is to add a number to it,
    I can validate that like so:

    >>> equivalence_validator = partial(
    ...     validate_codec,
    ...     command=lambda obj: obj + 1,
    ...     comparison=eq
    ... )
    >>> assert equivalence_validator(a)
    >>> assert equivalence_validator(b)


    So far, we've only used the default `pickle` encoder and decoders.
    Let's try out another pair.

    >>> import array
    >>> import struct
    >>> from functools import partial
    >>>
    >>> def float_encoder(a):
    ...     return struct.pack(len(a) * 'f', *a)
    ...
    >>> def float_decoder(b):
    ...     n = struct.calcsize('f')
    ...     return list(struct.unpack(int(len(b) / n) * 'f', b))
    >>>
    >>> a_list = [1, 2, 3]
    >>> a_tuple = (1, 2, 3)
    >>> an_array_of_floats = array.ArrayType('f', [1 / 100, 100])
    >>> an_array_of_doubles = array.ArrayType('d', [1 / 100, 100])
    >>>
    >>> float_validator = partial(
    ...     validate_codec,
    ...     coder=float_encoder,
    ...     decoder=float_decoder,
    ...     comparison=eq
    ... )
    >>>
    >>> assert float_validator(a_list)
    >>> assert not float_validator(a_tuple)
    >>> assert not float_validator(an_array_of_floats)

    The reason why `a_tuple` and `an_array_of_floats` don't validate is because
    array_codec decodes all iterables as lists, which are then compared to
    a tuple and an array (different types)
    Instead of using eq here, we should compare all elements individually:

    >>> all_equal = lambda x, y: all(xi == yi for xi, yi in zip(x, y))
    >>> float_validator_2 = partial(float_validator, comparison=all_equal)
    >>> assert float_validator_2(a_list)
    >>> assert float_validator_2(a_tuple)
    >>> assert float_validator_2(an_array_of_floats)
    >>> # but...
    >>> assert not float_validator_2(an_array_of_doubles)

    Why?
    Because we're using 'f' formatting spec, which is a float with 4 bytes
    yet the array is uses 'd' formatting (for doubles)
    It's easy to get into a... pickle of comparison comparing floats to doubles.
    We should probably use math.isclose to compare instead...

    >>> import math
    >>> all_are_close = lambda x, y: all(
    ...     math.isclose(xx, yy, rel_tol=1e-6) for xx, yy in zip(x, y)
    ... )
    >>>
    >>> float_validator_3 = partial(float_validator, comparison=all_are_close)
    >>> assert float_validator_3(a_list)
    >>> assert float_validator_3(a_tuple)
    >>> assert float_validator_3(an_array_of_floats)
    >>> assert float_validator_3(an_array_of_doubles)  # works with doubles now!

    """
    try:
        decoded_obj = encode_and_decode(obj, coder, decoder)
        obj_result = command(obj)
        decoded_obj_result = command(decoded_obj)
        return comparison(obj_result, decoded_obj_result)
    except caught_errors:
        return False



"""General python testing utils

A little tour...

Let's start with `validate_codec`, a function to test encoder/decoder pairs.

>>> from tested import validate_codec

`pickle.dumps/pickle.loads` is the default encoder/decoder pair.
You can pickle lists, and datetime objects

>>> validate_codec([1, 2, 3])
True
>>> from datetime import datetime
>>> validate_codec(datetime.now())
True

But you can't pickle a lambda function

>>> validate_codec(lambda x: x)
False

>>> from functools import partial
>>> import json
>>> validate_jsonability = partial(validate_codec, coder=json.dumps, decoder=json.loads)

You can jsonize lists and dicts

>>> assert validate_jsonability([1, 2, 3])
>>> assert validate_jsonability({'a': 1, 'b': {'c': [1, 2, 3]}})

You can't jsonize datetime objects

>>> from datetime import datetime
>>> validate_jsonability(datetime.now())
False

See `validate_codec` docs for more examples.

"""

from tested.multiple import run_multiple_pytests
from tested.ml import train_test_split_keys, learner_equivalence
from tested.codecs import encode_and_decode, validate_codec



"""
Utils for testing with local files.

Things like:

Making folder paths (and ensuring things about them):

>>> import os
>>> f = temp_dirpath()
>>> assert os.path.isdir(f)  # the directory exists (by default)
>>> assert dir_is_empty(f)  # ... and is empty (by default)

Making file paths (and ensuring things about them):

>>> filepath = temp_filepath()
>>> containing_dirpath = os.path.dirname(filepath)
>>> assert os.path.isdir(containing_dirpath)

"""
import tempfile
import shutil
import os
from typing import Optional
from pathlib import Path

file_sep = os.path.sep
DFLT_SUBPATH = 'tempdir'


def non_empty_tail(path):
    """The tail of the path, disregarding trailing slash if present.

    If / is the separator, this means:

    ```
        This/is/a/path -> path
        This/is/a/path/ -> path
    ```

    >>> path = os.path.join('This', 'is', 'a', 'path')
    >>> non_empty_tail(path)
    'path'
    >>> non_empty_tail(path + os.path.sep)
    'path'
    """
    if path.endswith(file_sep):
        path = path[:-1]
    return os.path.basename(path)


def empty_dir(dirpath):
    """Empty a directory of it's contents (and recreated the directory)
    Note: This means that the directory creation date will be now.
    """

    shutil.rmtree(dirpath)  # delete dir and all contents
    os.makedirs(dirpath)  # recreate the dir


def dir_is_empty(dirpath):
    """

    :param path:
    :return:

    >>> dirpath = temp_dirpath(ensure_is_empty=True)
    >>> dir_is_empty(dirpath)
    True
    """
    return not any(Path(dirpath).iterdir())


def temp_dirpath(
    subpath: str = DFLT_SUBPATH,
    ensure_exists: Optional[bool] = True,
    ensure_is_empty: Optional[bool] = True,
):
    """Get a fresh temporary folder path with assurances about existence and emptiness.

    :param subpath: The (relative) name of the folder.
    :param ensure_exists:
    :param ensure_is_empty:
    :return: A path (string) of a directory in a temporary


    >>> import os
    >>> from tested.local_files import temp_dirpath, non_empty_tail, dir_is_empty
    >>>
    >>> f = temp_dirpath()
    >>> assert non_empty_tail(f) == 'tempdir'
    >>> f = temp_dirpath('your_choice_of_a_dirname')
    >>> assert non_empty_tail(f) == 'your_choice_of_a_dirname'  # the directory name is indeed what you asked for
    >>> assert os.path.isdir(f)  # the directory exists!
    >>> assert dir_is_empty(f)  # ... and is empty
    >>> assert os.listdir(f) == []  # see!

    Let's write stuff in it:

    >>> import pathlib
    >>> p = pathlib.Path(f)
    >>> contents = 'hello world!'
    >>> assert p.joinpath('temp_file.txt').write_text(contents) == len(contents)
    >>> assert p.joinpath('temp_file.txt').read_text() == contents
    >>> assert os.listdir(f) == ['temp_file.txt']

    By default ``ensure_is_empty=True``, so you got an empty directory. But if you say False...

    >>> ff = temp_dirpath('your_choice_of_a_dirname', ensure_is_empty=False)
    >>> assert ff == f  # same path as before, but...
    >>> assert not dir_is_empty(f)
    >>> assert os.listdir(f) == ['temp_file.txt']
    >>>
    >>>
    >>> ff = temp_dirpath('your_choice_of_a_dirname', ensure_is_empty=True)
    >>> assert ff == f  # same path as before, but...
    >>> assert os.listdir(f) == []

    By default ``ensure_exists=True``, but the value could be:
    - None; meaning don't even check
    - False; meaning check, and if it exists, remove it

    >>> ff = temp_dirpath('your_choice_of_a_dirname', ensure_exists=None)
    >>> assert ff == f  # same path as before
    >>> assert os.path.isdir(ff)
    >>>
    >>> f = temp_dirpath('your_choice_of_a_dirname', ensure_exists=False)
    >>> assert not os.path.isdir(f)

    """
    dirpath = os.path.join(tempfile.gettempdir(), subpath)

    dir_exists = os.path.isdir(dirpath)
    if ensure_exists is True and not dir_exists:
        os.makedirs(dirpath)
    elif ensure_exists is False and dir_exists:
        os.removedirs(dirpath)
    if ensure_is_empty is not None:
        if ensure_is_empty and os.path.isdir(dirpath):
            shutil.rmtree(dirpath)  # delete dir and all contents
            os.makedirs(dirpath)  # recreate the dir
    return dirpath


def temp_filepath(
    filename='temp_file',
    subdir='temp_filepaths/',
    ensure_containing_dirs_exist=True,
    ensure_file_does_not_exist=False,
):
    """Make a temp filepath, ensuring (by default) that the containing directories exist,
    and (optionally) that the file doesn't exist either.

    >>> filepath = temp_filepath()
    >>> containing_dirpath = os.path.dirname(filepath)
    >>> assert os.path.isdir(containing_dirpath)
    >>> filepath = temp_filepath('my_own_name.txt')
    >>> assert os.path.basename(filepath) == 'my_own_name.txt'

    Let's write something in that file.

    >>> _ = Path(filepath).write_text('hello file!')  # but we can write in it
    >>> assert os.path.isfile(filepath)  # and now it exists
    >>> assert Path(filepath).read_text() == 'hello file!'  # here's what we wrote

    If you ask for that filepath again (a short time later), you'll get the same filepath,
    and the file will already exist. If you want to check if the file exists and delete it
    if it does, use ``ensure_file_does_not_exist==True``:

    >>> assert os.path.isfile(filepath)  # before
    >>> filepath2 = temp_filepath('my_own_name.txt', ensure_file_does_not_exist=True)
    >>> assert filepath2 == filepath
    >>> assert not os.path.isfile(filepath)  # after (it doesn't exist!)

    """
    dirpath = temp_dirpath(subdir, ensure_exists=True, ensure_is_empty=False)
    filepath = os.path.join(dirpath, filename)
    if ensure_containing_dirs_exist:
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
    if ensure_file_does_not_exist and os.path.isfile(filepath):
        os.remove(filepath)
    return filepath



"""Some objects to use in tests"""

from functools import partial
from operator import eq


class A:
    def __init__(self, x):
        self.x = x

    def __add__(self, other):
        return self.x + other


class B(A):
    def __eq__(self, other):
        return self.x == other.x


a_list = [1, 2, 3]
a_tuple = (1, 2, 3)
a = A(42)
assert a + 1 == 43
b = B(42)



"""
Testing utils for ML
"""

from typing import Iterable, Callable, Union, Tuple

import numpy as np
from sklearn.model_selection import GroupShuffleSplit
from sklearn.base import is_regressor, is_classifier, is_outlier_detector
from sklearn.datasets import make_classification, make_regression
from sklearn.base import BaseEstimator, TransformerMixin
from i2.signatures import call_forgivingly


# ---------------------------------------------------------------------------------------
# Learner equivalence

Learner = BaseEstimator  # but not necessarily fitted
Model = BaseEstimator  # but fitted
Estimator = Union[Callable, str, type, BaseEstimator]
XY = Tuple[Iterable, Iterable]
XYFactory = Callable[[], XY]


def learner_equivalence(
    learner_1, learner_2, xy=None, model_action=None, equivalence_scorer=None
):
    """Returns an score that measures how much the two learners are equivalent.
    The user can specify what data to use (`xy`) to fit the learners,
    what action to take on the fitted model (`model_action`),
    and what function to apply to the two results to compute the final score.

    But the user doesn't HAVE to specify all that usually (if the learners are
    all proper sklearn estimators) -- instead, the function will try to figure
    out defaults for any of these if not given.

    >>> from sklearn.linear_model import LinearRegression as Regressor
    >>> from sklearn.linear_model import RidgeClassifier as Classifier
    >>> from sklearn.decomposition import PCA as UnsupervisedTransformer
    >>>
    >>> from tested.ml import learner_equivalence
    >>>
    >>> for learner in (Regressor, Classifier, UnsupervisedTransformer):
    ...     # assert that a learner is equivalent to itself
    ...     assert learner_equivalence(learner, learner)
    """
    # preprocess inputs
    learner_1 = get_learner(learner_1)
    learner_2 = get_learner(learner_2)
    if xy is None:
        xy = get_xy_factory_for_estimator(learner_1)
    if isinstance(xy, Callable):
        X, y = xy()
    else:
        X, y = xy
    model_action = model_action or learner_1
    model_action, dflt_equivalence_scorer = get_model_action_and_equivalence_scorer(
        model_action
    )
    equivalence_scorer = equivalence_scorer or dflt_equivalence_scorer
    # do the stuff
    learner_1.fit(X, y)
    learner_2.fit(X, y)
    learner_1_output = model_action(learner_1, X)
    learner_2_output = model_action(learner_2, X)
    return equivalence_scorer(learner_1_output, learner_2_output)


def _is_estimator_factory(estimator) -> bool:
    """Tells us if the input might be able to be called to get an estimator"""
    return isinstance(estimator, type) or (
        isinstance(estimator, Callable) and not isinstance(estimator, BaseEstimator)
    )


valid_estimator_kinds = {'classifier', 'regressor', 'transformer'}


def estimator_kind(estimator: Estimator) -> str:
    """Returns the kind (string) of an Estimator"""
    if _is_estimator_factory(estimator):
        estimator = estimator()
    if isinstance(estimator, BaseEstimator):
        if is_classifier(estimator):
            kind = 'classifier'
        elif is_regressor(estimator):
            kind = 'regressor'
        elif TransformerMixin in type(estimator).mro():
            kind = 'transformer'
        else:
            kind = 'regressor'  # we'll just use that for unsupervised?
    elif isinstance(estimator, str):
        kind = estimator
    else:
        raise ValueError(f"Couldn't result estimator to a kind: {estimator}")
    return kind


def get_xy_factory_for_estimator(estimator: Estimator) -> XYFactory:
    kind = estimator_kind(estimator)

    if isinstance(kind, str):
        data_generators = {
            'classifier': make_classification,
            'regressor': make_regression,
            'transformer': make_regression,  # or classification better?
        }
        data_gen = data_generators.get(kind, None)
        if data_gen is None:
            raise ValueError(f'A string estimator must be one of {data_generators}')
    elif isinstance(kind, Callable):
        data_gen = kind
    else:
        raise ValueError(f'Unrecognized kind of estimator: {estimator}')
    return data_gen


def get_learner(learner) -> Learner:
    if isinstance(learner, type):
        # TODO: Do this for any Callable? (Need distinguish with callable instance)
        learner = learner()
    return learner


def get_model_action_and_equivalence_scorer(model_action):
    """Returns a default (model_action, equivalence_scorer) for a given model_action
    It will return the model_action as is if not a learner.
    If a learner, it will try to figure out a default model_action for it.
    """
    if isinstance(model_action, Learner):
        learner = model_action
        _estimator_kind = estimator_kind(learner)
        return {
            'classifier': (lambda model, X: getattr(model, 'predict')(X), np.allclose,),
            'regressor': (lambda model, X: getattr(model, 'predict')(X), np.allclose),
            'transformer': (
                lambda model, X: getattr(model, 'transform')(X),
                np.allclose,
            ),
        }.get(_estimator_kind)
    else:
        return model_action, np.allclose


# ---------------------------------------------------------------------------------------
# Train-test splits


def keys_aligned_list(iterable_spec, keys):
    """Get an iterable that is aligned with the keys iterable, and verify that it is so.

    >>> keys_aligned_list(lambda x: x * 2, keys=[1, 2, 3])
    [2, 4, 6]
    >>> keys_aligned_list([2, 4, 6], keys=[1, 2, 3])
    [2, 4, 6]
    >>> assert keys_aligned_list(None, keys=[1, 2, 3]) is None

    :param iterable_spec:
    :param keys:
    :return:
    """
    if iterable_spec is None:
        return None
    elif isinstance(iterable_spec, Callable):
        return list(map(iterable_spec, keys))
    elif isinstance(iterable_spec, Iterable):
        iterable_spec = list(iterable_spec)
        assert len(iterable_spec) == len(keys)
        return iterable_spec
    else:
        raise TypeError(
            f'Unknown iterable_spec type ({type(iterable_spec)}): {iterable_spec}'
        )


def train_test_split_keys(
    keys: Iterable,
    key_to_tag: Union[Callable, Iterable, None] = None,
    key_to_group: Union[Callable, Iterable, None] = None,
    *,
    test_size=None,
    train_size=None,
    random_state=None,
    n_splits=1,
):
    """Split keys into train and test lists.

    The ``train_keys`` and ``test_keys`` are disjoint and taken from ``keys``.

    Specifying ``key_to_tag`` (a function or iterable) ensures that ``tags`` will be
    well distributed in both train and test.

    Specifying ``key_to_group`` (a function or iterable) ensures **on the contrary**
    that keys of a same group will be entirely in train **or (exclusive)** in
    test -- not both.

    :param keys: keys to be split
    :param key_to_tag: keys-aligned iterable of tags (a.k.a y/classes in
        sklearn speak) or function to compute these from keys
    :param key_to_group: keys-aligned iterable of groups or function to compute
        these from keys
    :return a ``(train_keys, test_keys)`` pair (all elements of keys) if ``n_splits=1``,
        and a generator of such pairs if not.

    Note that in the doctest below, we take ``keys=[7, 14, 21, ...]`` to show that
    it's not about ``[0, 1, 2, ...]`` indices only, but ANY keys
    (even non numerical -- like filepaths, DB selectors, etc.)

    >>> keys = range(7, 7 + 100 * 7, 7)  # [7, 14, 21, ..., 700]
    >>> def mod5(x):
    ...     return x % 5
    >>> train_keys, test_keys = train_test_split_keys(keys, key_to_group=mod5,
    ...     train_size=.5, random_state=42)

    Observe here that though ``train_size=.5``, the proportion is not 50/50.
    That's because the group constraint, imposed by the key_to_group argument
    produces only 5 groups.

    >>> len(train_keys), len(test_keys)
    (40, 60)

    But especially, see that though there's a lot of train and test indices,
    within train, there's only 2 unique groups (all 0 or 3 modulo 5)
    and only 3 unique groups (1, 2, 4 modulo 5) within test indices.

    >>> assert set(map(mod5, train_keys)) == {0, 3}
    >>> assert set(map(mod5, test_keys)) == {1, 2, 4}

    """
    splitter = call_forgivingly(
        GroupShuffleSplit, **locals()
    )  # calls GroupShuffleSplit on relevant inputs

    keys = np.array(list(keys))
    y = keys_aligned_list(key_to_tag, keys)
    groups = keys_aligned_list(key_to_group, keys)
    if groups is None:
        groups = range(len(keys))

    n = splitter.get_n_splits(keys, y, groups)
    if n == 1:
        train_idx, test_idx = next(splitter.split(keys, y, groups))
        return keys[train_idx], keys[test_idx]
    else:

        def gen():
            for train_idx, test_idx in splitter.split(keys, y, groups):
                yield keys[train_idx], keys[test_idx]

        return gen()



"""Tools to contextualize tests

>>> versions()  #doctest: +SKIP
['system',
 '3.8.6',
 '3.8.6/envs/p3',
 '3.8.6/envs/test',
 '3.9.0',
 '3.9.0/envs/p39',
 '3.9.0/envs/test2',
 '* p3 (set by PYENV_VERSION environment variable)',
 'p39',
 'test',
 'test2']

>>> t = versions_available_for_install()   #doctest: +SKIP
>>> print(f"{len(t)=}")   #doctest: +SKIP
len(t)=436
>>> t[:5]   #doctest: +SKIP
['2.1.3', '2.2.3', '2.3.7', '2.4.0', '2.4.1']

See:
https://stackoverflow.com/questions/70177366/context-manager-to-setup-and-tear-down-test-virtual-environments

"""

import subprocess


def syscall(*args, decoder=bytes.decode, split_single_strings=False, post_proc=None):
    if decoder is None:
        decoder = lambda x: x
    if split_single_strings and len(args) == 1 and isinstance(args[0], str):
        single_string = args[0]
        args = single_string.split(' ')
    result = subprocess.run(args, capture_output=True)
    if result.returncode:
        raise RuntimeError(f'Error {result.returncode}: {result.stderr.decode()}')
    else:
        decoded_stdout = decoder(result.stdout)
        if not post_proc:
            return decoded_stdout
        else:
            return post_proc(decoded_stdout)


def lines_string_to_list(lines_string, skip_lines=0, line_split_char='\n'):
    x = list(filter(None, map(str.strip, lines_string.split(line_split_char))))
    return x[skip_lines:]


from functools import partial

print_versions = partial(syscall, 'pyenv', 'versions', post_proc=print)
versions = partial(syscall, 'pyenv', 'versions', post_proc=lines_string_to_list)
versions_available_for_install = partial(
    syscall,
    'pyenv',
    'install',
    '--list',
    post_proc=partial(lines_string_to_list, skip_lines=1),
)
mk_virtual_env = partial(syscall, 'pyenv', 'virtualenv')



"""Tools for testing with notebooks.

We often test our stuff somewhere in a notebook.

Since we already spent time doing that, how can we get more out of the investment
by making our work more reusable?

Ideas.

Gather the relevant cells and generate a string that can be pasted into a test
 module/function. Problems:
- Getting the right environment (imports, data, etc.)
- asserting (can use out cells sometimes, but sometimes harder)

For doctests? Change comments to markdown text and code to doctests.

Links to check out:
- https://semaphoreci.com/blog/test-jupyter-notebooks-with-pytest-and-nbmake

"""

from itertools import starmap
from typing import Iterable, Any


class CellFuncs:
    """A collection of cell funcs to use with InOut.cells_2_str"""

    @staticmethod
    def out_as_comment(in_, out_, prefix='# Out: '):
        """Insert out_ cell as comment after in_"""
        if out_ is not None:
            out_str = str(out_)
            if '\n' not in out_str:
                out_str = prefix + out_str
            else:
                out_str = (
                    prefix + '\n' + '\n'.join(('# ' + x for x in out_str.split('\n')))
                )
            return in_ + '\n' + out_str
        else:
            return in_

    @staticmethod
    def assertion_when_it_works(
        in_,
        out_,
        fallback=lambda in_, out_: in_,  # just return None
        globals_=None,
        locals_=None,
    ):
        r"""
        >>> CellFuncs.assertion_when_it_works("3 + 4", 7)
        'assert (3 + 4) == (7)\n'
        >>> CellFuncs.assertion_when_it_works("dict(a=1, b=2)", {'a': 1, 'b': 2})
        "assert (dict(a=1, b=2)) == ({'a': 1, 'b': 2})\n"

        And if the left and right are not equal when evaluated,
        by default (fallback) just the in_ is returned

        >>> CellFuncs.assertion_when_it_works("dict(a=1, b=2)", "is not equal to this")
        'dict(a=1, b=2)'

        """
        if out_ is not None:
            command = f'({in_}) == ({out_})\n'
            try:
                bool_ = eval(command, globals_, locals_)
            except Exception:
                bool_ = False
            if bool_:
                return f'assert {command}'
            else:
                return fallback(in_, out_)
        else:
            return in_


InCell = str
OutCell = Any

from typing import Protocol, runtime_checkable
from i2 import Sig


@runtime_checkable
class InCellFilter(Protocol):
    def __call__(self, in_cell: str) -> bool:
        pass


@runtime_checkable
class InOutCellFilter(Protocol):
    def __call__(self, in_cell: str, out_cell: Any) -> bool:
        pass


class InOut:
    """Get (in, out) pairs of cell contents.

    In a notebook, do:
    ```
    t = InOut(In, Out)
    ```

    `In` and `Out` are automatically available in a notebook.

    Now you can access (in, out) pairs with `t[k]` where `k` can be an integer, a slice,
    or a list/iterable of integers, or a (in, or (in, out) filter) function.

    Examples:

    ```
    t[4]  # the (in, out) cell for idx=4

    print(t.cells_2_str(t[10:20])

    print(t.cells_2_str(t[[77, 86, 99, 87, 89]],
            cell_func=CellFuncs.assertion_when_it_works))

    # number of in_ cells that have exactly one line
    len(t[lambda in_: len(in_.split('\n')) == 1])

    # number of out cells that are not None
    len(t[lambda in_, out_: out_ is not None])
    ```

    """

    def __init__(self, In, Out):
        self.In = In
        self.Out = Out

    def __getitem__(self, k):
        if isinstance(k, int):
            return self.In[k], self.Out[k]
        else:
            return [
                (self.In[idx], self.Out[idx] if idx in self.Out else None)
                for idx in self.idx_gen(k)
            ]

    def _in_out_pairs_with_sentinel_out_when_missing(self, sentinel=None):
        for i, in_ in enumerate(self.In):
            if i in self.Out:
                yield in_, self.Out[i]
            else:
                yield in_, sentinel

    def idx_gen(self, k):
        if isinstance(k, int):
            yield k
        elif isinstance(k, slice):
            yield from range(k.start, k.stop, k.step or 1)
        elif isinstance(k, (InCellFilter, InOutCellFilter)):
            # Note, the above ininstance check is just a "is callable" check!
            n_required_args = Sig(k).n_required
            if n_required_args == 1:
                in_filt = k
                yield from (i for i, in_ in enumerate(self.In) if in_filt(in_))
            elif n_required_args == 2:
                in_out_filt = k
                in_outs = self._in_out_pairs_with_sentinel_out_when_missing()
                yield from (
                    i for i, (in_, out_) in enumerate(in_outs) if in_out_filt(in_, out_)
                )
            else:
                raise ValueError(f'A cell filter must have 1 or 2 required arguments')
        elif isinstance(k, Iterable):
            yield from k

    def index_has_an_in(self, idx):
        return idx in self.In

    @staticmethod
    def cells_2_str(cells_list, cell_func=CellFuncs.out_as_comment, sep='\n'):
        """Get a string out of a list of cells.
        Probably the main utility of InOut.
        """
        return sep.join(starmap(cell_func, cells_list))

    @classmethod
    def from_locals(cls, _locals):
        """A convenience function to get an InOut.

        Pretty much only one way to call it, from your notebook, do this:

        ```
        t = InOut(locals())
        ```

        """
        return cls(_locals['In'], _locals['Out'])



"""Utils to test multiple modules and dependencies"""

import os
import subprocess


def ensure_path(module):
    if isinstance(module, str) and os.path.exists(module):
        return module  # it is a path
    else:  # assume it's a module object, and get the path for it
        path = (
            module.__file__
        )  # TODO: Use other means, since __file__ doesn't work for some packages
        if path.endswith('__init__.py'):  # if path is an __init__.py
            path = os.path.dirname(path)  # ... take the containing folder
        return path


def multiple_pytests_commands(
    *modules_to_test, options: str = '--doctest-modules --quiet --disable-warnings'
):
    template = 'pytest ' + options + ' {path}; '
    for path in map(ensure_path, modules_to_test):
        yield template.format(path=path)


def multiple_pytests_command_str(
    *modules_to_test, options: str = '--doctest-modules --quiet --disable-warnings'
):
    return ''.join(multiple_pytests_commands(*modules_to_test, options=options))


def run_multiple_pytests(
    *modules_to_test,
    options: str = '--doctest-modules --quiet --disable-warnings',
    dry_run=False
):
    for command in multiple_pytests_commands(*modules_to_test, options=options):
        print(command)
        if not dry_run:
            r = subprocess.check_output(command, shell=True)
            if isinstance(r, bytes):
                r = r.decode()
            print(r.strip() + '\n')
